# -*- coding: utf8 -*-
"""
Logging decorator that logs function name, args and kwargs.

Usage:
    from decorators import log_function_io

    @log_function_io()
    def foo(bar, spam=None, *args, **kwargs):
        return do_stuff()

    >>> foo('a', *['first star', 'second star'], **{'something': 'boo'})
    2016-02-02 05:04:50,963 - __main__ - DEBUG - [FUN] foo [ARG] bar: 'a', spam: 'first star', something: 'boo' *('second star',)
    2016-02-02 05:04:50,963 - __main__ - DEBUG - [FUN] foo [RET] None
"""
from collections import OrderedDict
from functools import wraps
import functools
import inspect
import itertools
import logging
import sys
import time


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class CustomAdapter(logging.LoggerAdapter):

    """ Custom LoggerAdapter that adds metadata to the log message. """

    skip_processing_tags = ('[ARG]', '[RET]', '[FUN]')
    skip_msg_tag_tags = ('[POST]', '[GET]')

    def process(self, msg, kwargs):
        """Adds additional metadata to log message.

        Takes passed log message and adds additional metadata, like function
        name and [FUN], [MSG] tags.

        If `msg` already has tags specified in `skip_processing_tags`, it
        returns original message.
        If `msg` has tag from `skip_msg_tag_tags`, it only adds [FUN] tag and
        skips adding [MSG] tag.

        :param str msg: Log string to process
        :param dict kwargs: Keyword arguments for log message
        :return: Processed message and original keyword arguments
        :rtype: tuple
        """
        def get_func_name():
            stack = inspect.stack()
            for index, frame in enumerate(stack):
                if frame[3] == 'process':
                    return stack[index + 2][3]

        if any(tag in msg for tag in self.skip_processing_tags):
            return msg, kwargs
        elif any(tag in msg for tag in self.skip_msg_tag_tags):
            msg = u'[FUN] {0} {1}'.format(get_func_name(), msg)
        else:
            msg = u'[FUN] {0} [MSG] {1}'.format(get_func_name(), msg)

        return msg, kwargs


def get_logger_adapter(name):
    """Returns Custom LoggerAdapter that adds '[MSG]' tag to log message.

    Function returns standard logger, if name of the module is
    'log_decorator', because we don't wont to get this tag in
    `log_function_io` decorator.

    :param str name: Name of the module, for which to get logger
    :return: Logger adapter for given module
    :rtype: CustomAdapter
    """
    logger = logging.getLogger(name)
    return CustomAdapter(logger, {})


def log_function_io(f):

    log = getattr(
        sys.modules[f.__module__],
        'log',
        get_logger_adapter(f.__module__)
    )

    @wraps(f)
    def wrapper(*args, **kwargs):
        passed_args = inspect.getargspec(f)[0]
        args_name = list(
            OrderedDict.fromkeys(passed_args + kwargs.keys())
        )
        args_dict = OrderedDict(
            list(itertools.izip(args_name, args)) +
            list(kwargs.iteritems())
        )

        star_args = []
        if args:
            star_args = args[len(passed_args):]

        # We don't care about `self` or `cls`
        args_dict.pop('self', None)
        args_dict.pop('cls', None)

        # Build a log_message with argument name and value
        log_args_message = ', '.join(
            '{0}: %r'.format(arg_name.encode('utf-8'))
            for arg_name in args_dict.keys()
        )

        if log_args_message and not star_args:
            log_args_message = '[FUN] {0} [ARG] {1} '.format(
                f.func_name,
                log_args_message
            )
        elif log_args_message and star_args:
            log_args_message = '[FUN] {0} [ARG] {1} *{2}'.format(
                f.func_name,
                log_args_message,
                star_args
            )
        else:
            log_args_message = '[FUN] {0} [ARG] No arguments'.format(
                f.func_name
            )

        log.debug(log_args_message, *args_dict.values())

        return_value = f(*args, **kwargs)

        log_return_value_message = '[FUN] {0} [RET] {1}'.format(
            f.func_name,
            repr(return_value).decode('utf-8')
        )

        log.debug(log_return_value_message)

        return return_value

    return wrapper



# -*- coding: utf8 -*-
"""
A decorator to profile function calls.

Usage:
    from mj_helpers.decorator import profileit

    @profileit
    def foo():
        return do_stuff()

In a Python shell:
    >>> form pstats import Stats
    >>> stats = Stats('/tmp/foo.profile')
    >>> stats.sort_stats('cumulative').print_stats(50)

"""
try:
    import cProfile as profile
except ImportError:
    import profile

def profileit(func):

    def wrapper(*args, **kwargs):
        datafn = '/tmp/{0}.profile'.format(func.__name__)
        prof = profile.Profile()
        retval = prof.runcall(func, *args, **kwargs)
        prof.dump_stats(datafn)

        return retval

    return wrapper



# -*- coding: utf-8 -*-
"""
Caching decorator to assist in caching function calls / returns.

Usage:
    from mj_helpers.decorators import cache_it

    @cache_it()
    def foo(bar, spam=None, *args, **kwargs):
        return do_stuff()

    # If using memoize instead of django's cache you can see the cache by:
    >>> cache_it._cache
    {
        <function foo at 0x7f4419daa848>: {
            "(('a', 'first star', 'second star'), (('something', 'boo'),))":
            (None, 1454391947.614329)
        }
    }
"""
from functools import wraps
import inspect
import logging
import threading

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


try:
    from django.core.cache import cache

    def cache_it(cache_time=600, custom_prefix=None):

        """ Caches the data for ``cache_time`` seconds. """

        __cache_lock = threading.Lock()

        def decorator(func):

            @wraps(func)
            def inner(*args, **kwargs):
                # Little simplification to make same calls having same
                # cache_key by removing Nones and default values.
                # Also .defaults can be None, in such case we need to change
                # it into empty list for zip() to work
                func_args = inspect.getargspec(func)
                default_args = dict(
                    zip(
                        reversed(func_args.args or []),
                        reversed(func_args.defaults or [])
                    )
                )
                cache_key_kwargs = {
                    key: val for key, val in kwargs.iteritems()
                    if not (val is None or val == default_args.get(key))
                }
                cache_key = repr(func) + repr(args) + repr(cache_key_kwargs)

                # Redis key can not contain non ascii chars (cache_key is
                # always str, because of using repr)
                cache_key = cache_key.decode('utf-8')
                if custom_prefix:
                    cache_key = custom_prefix + cache_key

                with __cache_lock:
                    if cache_key not in cache:
                        result = func(*args, **kwargs)
                        cache.set(cache_key, result, cache_time)
                        return result

                    return cache.get(cache_key)

            return inner

        return decorator

except ImportError as err:
    logger.exception("Couldn't import Django, defaulting to memoize.")
    import time

    class cache_it(object):

        """ Memoize With Timeout.

        Borrowed with love from:
        http://code.activestate.com/recipes/325905-memoize-decorator-with-timeout/
        """
        _caches = {}
        _timeouts = {}

        def __init__(self, cache_time=600, custom_prefix=None):
            self.timeout = cache_time
            self.custom_prefix = custom_prefix

        def collect(self):
            """ Clear cache of results which have timed out. """
            for func in self._caches:
                cache = {}
                for key in self._caches[func]:
                    time_ = time.time() - self._caches[func][key][1]
                    if time_ < self._timeouts[func]:
                        cache[key] = self._caches[func][key]

                self._caches[func] = cache

        def __call__(self, function):
            self.cache = self._caches[function] = {}
            self._timeouts[function] = self.timeout

            def func(*args, **kwargs):
                keys = sorted(kwargs.items())
                key = repr((args, tuple(keys)))
                if self.custom_prefix:
                    key = repr(self.custom_prefix) + key

                try:
                    value = self.cache[key]
                    if (time.time() - value[1]) > self.timeout:
                        raise KeyError

                except KeyError:
                    value = self.cache[key] = function(*args, **kwargs), time.time()

                return value[0]
            func.func_name = function.__name__

            return func



#!/usr/bin/env python
# -*- coding: utf8 -*-
"""
Command line argument to print stats to screen after profileit has been run.
"""
from pstats import Stats
import argparse
import sys
import os


SORTING = {
    'calls': 'Call count',
    'cumulative': 'Cumulative time',
    'file': 'File name',
    'module': 'File name',
    'pcalls': 'Primitive call count',
    'line': 'Line number',
    'name': 'Function name',
    'nfl': 'Name/File/Line',
    'stdname': 'Standard name',
    'time': 'Internal time',
}

def print_stats(filepaths, sorting='cumulative', limit=50):
    """
    Print the stats of a the given files.

    Loop through filepaths, if the file exists, attempt to parse the file with
    Stats the print the results.

    :param filepaths: List of strings that represents file paths
    :param sorting: A string representing which field to sort on
    :param limit: The number of lines to print
    """
    for file_ in filepaths:
        if os.path.isfile(file_):
            try:
                print('\n')
                stats = Stats(file_)

            except AttributeError:
                error_message = '{file_} does not appear to be a valid file'
                print(error_message.format(file_=file_))
                continue

            stats.sort_stats('cumulative').print_stats(limit)

def print_sorting_options():
    print('\nSorting option must be one of the following:\n')
    table_break = '+-----------+---------------------+'
    print(table_break)
    print('|{key:<11}|{value:<21}|'.format(key='Valid Arg', value='Meaning'))
    print(table_break.replace('-', '='))
    for key, value in SORTING.iteritems():
        print('|{key:<11}|{value:<21}|'.format(key=key, value=value))
        print(table_break)


def main():
    description = 'Print the top 50 stats from the @profileit decorator'
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        'filepath',
        nargs='+',
        help='Location of /file/path/$FUNCTION_NAME.profile'
    )
    parser.add_argument(
        '--sorting',
        default='cumulative',
        help='Optional argument for sorting results'
    )
    parser.add_argument(
        '--limit',
        type=int,
        default=50,
        help='Optional argument for sorting results'
    )

    args = parser.parse_args()

    if args.sorting not in SORTING:
        print_sorting_options()
        sys.exit(-1)

    if args.filepath:
        print_stats(args.filepath, args.sorting, args.limit)


if __name__ == '__main__':
    main()



# -*- coding: utf8 -*-

from codecs import open
from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as file_:
    long_description = file_.read()

setup(
    name='mj_helpers',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.1',

    description=(
        'A collection of helper functions and decorators that I occasionally '
        'use to get my life in order.'
    ),
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/mikejarrett/mj-helpers',

    author='Mike Jarrett',
    author_email='mike.d.jarrett at gmail dot com',

    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    keywords='logging profiling decorators',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    install_requires=[],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={},

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'mj_stats=mj_helpers.mj_stats:main',
        ],
    },
)
